
//========================================================================================
//
//		Fun with prime numbers - the sieve of Eratosthenes v1
//		by Dan Brown
//		www.tacticaltinkering.co.uk
//
//========================================================================================

#include <iostream>
#include <Windows.h>

#define MAX		15485863			//15485863	//179424673

#if MAX > 4294967295				// If we can, use 32 bit integers (twice as fast on a 32 bit system)
typedef unsigned __int64 BigInt;
#else
typedef unsigned int BigInt;
#endif

using namespace std;

//-----------------------------------------------------------------------------------------
// Name: CTimer [CLASS]
// Desc: A class for use as a fairly accurate stopwatch
//-----------------------------------------------------------------------------------------
class CTimer	{
	private:
		__int64 start;
		__int64 end;
		__int64 iElapsed()
		{
			LARGE_INTEGER lInt;
			QueryPerformanceCounter(&lInt);
			return (lInt.QuadPart-start);
		}
	public:
		double value;
		double valueMS;
		bool Start()
		{
			LARGE_INTEGER lInt;
			if(!QueryPerformanceCounter(&lInt))
			{
				std::cout << "\nTimer failed to start!";
				return false;
			}
			start = lInt.QuadPart;
			return true;
		}
		bool Stop()
		{
			LARGE_INTEGER lIntPF, lInt;
			if(!QueryPerformanceFrequency(&lIntPF)) return false;
			if(!QueryPerformanceCounter(&lInt)) return false;			
			end = lInt.QuadPart-start;
			value = (double)iElapsed()/(double)lIntPF.QuadPart;
			valueMS = value*1000;
			return true;
		}
		double TimeElapsed()
		{
			Stop();
			return value;
		}
} Timer;
//-----------------------------------------------------------------------------------------

char *n = new char[MAX];	// Create the main array on the heap

int main()
{
	Timer.Start();

	// Assume all numbers (bar 1) are prime to start, so set all array values to 1 except the first
	std::fill(n, n + MAX, 1);
	n[0] = 0;

	cout << "\nInitialised the array (" << Timer.TimeElapsed() << "s)\n\nGenerating primes with the sieve of Eratosthenes ...\n\n";

	int p = 2;				// p = current prime we're deleting multiples of in the bit array
	int m = p*p;			// m = multiples of p greater than (p-1)*p

	while(m <= MAX)
	{
		while(m <= MAX)		// Loop through the array and remove all multiples of p
		{
			n[m-1] = 0;
			m += p;
		}
		
		do {p++;} while( n[p-1] != 1 );		// Let p be the next surviving number in the list
		m=p*p;								// Starting multiple of p
	}

	Timer.Stop();
	
	BigInt primesFound = 0;
	for( BigInt a = 1 ; a <= MAX ; a++ )
		if(n[a-1]==1) 
		{
			primesFound++; 

			if(primesFound==1000000) cout << "\nMillionth prime: " << a;
			if(primesFound==10000000) cout << "\n10 millionth prime: " << a;
			if(primesFound==50000000) cout << "\n50 millionth prime: " << a;
			if(primesFound==100000000) cout << "\n100 millionth prime:\t" << a;
		}

	cout << "\n\nPrimes found: " << primesFound << "\nSearched up to: " << MAX << "\n\nTime elapsed: ";
	Timer.value > 1 ? cout << Timer.value << "s" : cout << Timer.valueMS << "ms";
	
	cin.ignore();

	delete [] n;
}
